-- vertex aware spritebatch for spritebatch mesh instancing
VASpriteBatch = {}
VASpriteBatch.__index = VASpriteBatch
VASpriteBatch.ident = "c_vaspritebatch"
VASpriteBatch.use_vertices = true
VASpriteBatch.use_lazy_autorefresh = true
VASpriteBatch.ar_position = true -- always should be false until i understand how it works
VASpriteBatch.ar_texcoord = true -- same
VASpriteBatch.ar_color = true
function VASpriteBatch:new(texture,buffer_size,hint)
	local self = setmetatable({},VASpriteBatch)
	self.sb = love.graphics.newSpriteBatch(texture,buffer_size,hint)
	return self
end

function VASpriteBatch:init()
	if self.use_vertices == true then
		self.mesh = love.graphics.newMesh(4*self.sb:getBufferSize())
		self.vert_array = {}
	end
end

local sin,cos = math.sin,math.cos

function VASpriteBatch:add(quad, x, y, r, sx, sy, ox, oy, kx, ky)
	local id = self.sb:add(quad, x, y, r, sx, sy, ox, oy, kx, ky)
	if self.use_vertices == true then
		-- if we use separate vertex array (an by default we do)
		-- at the same time the sprite is created, we create
		-- vertex array which resembles our new sprite
		-- and we need co calculate each vertex data from sprite creation parameters
		-- for later transformations
		local qx,qy,qw,qh = quad:getViewport()
		local texw, texh = self.sb:getTexture():getWidth(), self.sb:getTexture():getHeight()
		-- calculate vertex position
		-- should include rotation, skew and offset
		local ox = ox or 0
		local oy = oy or 0
		local r = r or 0
		local sx = sx or 1
		local sy = sy or 1
		local vx = x - ox*sx
		local vy = y - oy*sy
		-- calculate vertex coordinates
		local tlx,tly = (vx),(vy)
		local blx,bly = (vx),(vy+qh*sy)
		local trx,try = (vx+qw*sx),(vy)
		local brx,bry = (vx+qw*sx),(vy+qh*sy)
		-- apply rotation
		local ntlx,ntly = x + (tlx - x)*cos(r) - (tly - y)*sin(r), y + (tlx - x)*sin(r) + (tly - y)*cos(r)
		local nblx,nbly = x + (blx - x)*cos(r) - (bly - y)*sin(r), y + (blx - x)*sin(r) + (bly - y)*cos(r)
		local ntrx,ntry = x + (trx - x)*cos(r) - (try - y)*sin(r), y + (trx - x)*sin(r) + (try - y)*cos(r)
		local nbrx,nbry = x + (brx - x)*cos(r) - (bry - y)*sin(r), y + (brx - x)*sin(r) + (bry - y)*cos(r)
		-- calculate vertex texture coordinates
		-- should include texture size and quad viewport size		
		local qx_n,qy_n = qx/texw,qy/texh
		local qw_n,qh_n = qw/texw,qh/texh
		-- add vertices for later transformations
		--[[ Sprite vertex numbers are as follows:
			  1-----3
			  |   / |
			  |  /  |
			  | /   |
			  2-----4 ]]--
		
		-- on each set or add sprite we add or recalculate vertices for our internal mesh
		-- for later attribute attachment
		-- top left vertex
		self.vert_array[#self.vert_array+1] = {ntlx,ntly, 	qx_n, qy_n, 		255,255,255,255}
		-- bottom left
		self.vert_array[#self.vert_array+1] = {nblx,nbly,	qx_n,qy_n+qh_n, 	255,255,255,255}
		-- top right
		self.vert_array[#self.vert_array+1] = {ntrx,ntry, 	qx_n+qw_n,qy_n, 	255,255,255,255} 
		-- bottom right
		self.vert_array[#self.vert_array+1] = {nbrx,nbry,	 qx_n+qw_n,qy_n+qh_n,255,255,255,255}
		self.mesh:setVertices(self.vert_array)
	end
	return id
end

function VASpriteBatch:clear() self.sb:clear() end
function VASpriteBatch:set(id,quad,x,y,r,sx,sy,ox,oy,kx,ky)	
	self.sb:set(id,quad,x,y,r,sx,sy,ox,oy,kx,ky) 
	if self.use_vertices == true then
	
		local qx,qy,qw,qh = quad:getViewport()
		local texw, texh = self.sb:getTexture():getWidth(), self.sb:getTexture():getHeight()
		local ox = ox or 0
		local oy = oy or 0
		local r = r or 0
		local sx = sx or 1
		local sy = sy or 1
		local vx = x - ox*sx
		local vy = y - oy*sy
		local tlx,tly = (vx),(vy)
		local blx,bly = (vx),(vy+qh*sy)
		local trx,try = (vx+qw*sx),(vy)
		local brx,bry = (vx+qw*sx),(vy+qh*sy)
		local ntlx,ntly = x + (tlx - x)*cos(r) - (tly - y)*sin(r), y + (tlx - x)*sin(r) + (tly - y)*cos(r)
		local nblx,nbly = x + (blx - x)*cos(r) - (bly - y)*sin(r), y + (blx - x)*sin(r) + (bly - y)*cos(r)
		local ntrx,ntry = x + (trx - x)*cos(r) - (try - y)*sin(r), y + (trx - x)*sin(r) + (try - y)*cos(r)
		local nbrx,nbry = x + (brx - x)*cos(r) - (bry - y)*sin(r), y + (brx - x)*sin(r) + (bry - y)*cos(r)		
		local qx_n,qy_n = qx/texw,qy/texh
		local qw_n,qh_n = qw/texw,qh/texh
		
		-- instead of creating new vertices, we change existing ones
		local vi = (id-1)*4
		self.vert_array[vi+1] = {ntlx,ntly, 	qx_n, qy_n, 		self.vert_array[vi+1][5],self.vert_array[vi+1][6],self.vert_array[vi+1][7],self.vert_array[vi+1][8]}
		self.vert_array[vi+2] = {nblx,nbly,		qx_n,qy_n+qh_n, 	self.vert_array[vi+2][5],self.vert_array[vi+2][6],self.vert_array[vi+2][7],self.vert_array[vi+2][8]}
		self.vert_array[vi+3] = {ntrx,ntry, 	qx_n+qw_n,qy_n, 	self.vert_array[vi+3][5],self.vert_array[vi+3][6],self.vert_array[vi+3][7],self.vert_array[vi+3][8]} 
		self.vert_array[vi+4] = {nbrx,nbry,	 	qx_n+qw_n,qy_n+qh_n,self.vert_array[vi+4][5],self.vert_array[vi+4][6],self.vert_array[vi+4][7],self.vert_array[vi+4][8]}
		self.mesh:setVertices(self.vert_array)
	end
end
function VASpriteBatch:swapSpriteGeometry(id1,id2)
	local vi_1 = ((id1-1)*4)
	local vi_2 = ((id2-1)*4)
	for i=1,4 do
		local vd_1 = self.vert_array[vi_1+i]
		local vd_2 = self.vert_array[vi_2+i]
		for j = 1,8 do
			vd_1[j],vd_2[j] = vd_2[j],vd_1[j]
		end
	end
	self.mesh:setVertices(self.vert_array)
end
function VASpriteBatch:setTexture(t) self.sb:setTexture(t) end
function VASpriteBatch:flush() self.sb:flush() end
function VASpriteBatch:getBufferSize() return self.sb:getBufferSize() end
function VASpriteBatch:setBufferSize(bs) self.sb:setBufferSize(bs) end
function VASpriteBatch:getTexture()	return self.sb:getTexture() end
function VASpriteBatch:attachAttribute(name,mesh) self.sb:attachAttribute(name,mesh or self.mesh) end
function VASpriteBatch:refreshAttributes(attr) self:attachAttribute(attr,self.mesh) end
function VASpriteBatch:getCount() return self.sb:getCount() end
function VASpriteBatch:setVertex(index,vdata) self.vert_array[index] = vdata end
function VASpriteBatch:getVertexData(index)	return self.vert_array[index] end
function VASpriteBatch:refreshVertexData() self.mesh:setVertices(self.vert_array) end

-- this function returns spritebatch to draw in love.draw
-- and to reduce attachAttributes call count to one before actual drawcall
-- i use lazy autorefresh of each mesh attribute before that drawcall, 
-- not after attribute has been modified
function VASpriteBatch:getSpriteBatch() 
	if self.use_lazy_autorefresh == true then
		self:refreshVertexData()
		if self.ar_position == true then self:refreshAttributes("VertexPosition") end
		if self.ar_texcoord == true then self:refreshAttributes("VertexTexCoord") end
		if self.ar_color == true then self:refreshAttributes("VertexColor") end
	end
	return self.sb 
end